STRUCTURE = "structure"
CONTINUOUS = "continuous"
SOURCE_PROCESSOR_NAME = "source_processor_name"
SOURCE_PROCESSOR_ID = "source_processor_id"
TRAILING_NUMBER = ".0"
SAMPLE_RATE = "sample_rate"
CHANNEL_NUMBER = "num_channels"



from functools import partial
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from numpy.typing import ArrayLike
from scipy.signal import savgol_filter

from neurokin.utils.features_extraction import feature_extraction
from neurokin.utils.kinematics import binning
from neurokin.utils.helper import load_config
from neurokin.utils.kinematics import kinematics_processing, import_export, event_detection

class KinematicDataRun:
    """
    This class represents the kinematics data recorded in a single run.
    """

    def __init__(self, path, configpath):
        self.path = path

        self.config = load_config.read_config(configpath)

        self.trial_roi_start: int
        self.trial_roi_end: int
        self.fs: float
        self.condition: str

        self.markers_df = pd.DataFrame()
        self.gait_param = pd.DataFrame()
        self.stepwise_gait_features = pd.DataFrame()
        self.features_df: pd.MultiIndex = None
        self.binned_df: pd.DataFrame = None

        self.left_mtp_lift: ArrayLike = None
        self.left_mtp_land: ArrayLike = None
        self.left_mtp_max: ArrayLike = None
        self.right_mtp_lift: ArrayLike = None
        self.right_mtp_land: ArrayLike = None
        self.right_mtp_max: ArrayLike = None
        self.bodyparts: ArrayLike = None
        self.scorer: str = "scorer"

    def load_kinematics(self,
                        correct_shift: bool = False,
                        correct_tilt: bool = False,
                        to_shift: ArrayLike = None,
                        to_tilt: ArrayLike = None,
                        shift_reference_marker: str = "",
                        tilt_reference_marker: str = "",
                        source: str = None,
                        fs: float = None):
        """
        Loads the kinematics from a c3d file into a dataframe with timeframes as rows and markers as columns

        :param correct_shift: bool should there be a correction in the shift of one of the axis?
        :param correct_tilt: bool should there be a correction in the tilt (linear trend) of one of the axis?
        :param to_shift: which columns to perform the shift on if correct_shift is true
        :param to_tilt: which columns to perform the tilt on if correct_shift is true
        :param shift_reference_marker: which marker to use as a reference trajectory to compute the shift
        :param tilt_reference_marker: which marker to use as a reference trajectory to compute the tilt
        :param source: defines which source of kinematics data to load from, either c3d or dlc
        :return:
        """

        if source.lower() == "c3d":
            self.trial_roi_start, self.trial_roi_end, self.fs, self.markers_df = import_export.import_c3d(self.path)
        elif source.lower() == "dlc":
            self.markers_df = import_export.import_dlc_df(self.path)
            self.convert_DLC_like_to_df()
            self.fs = fs
        else:
            raise ValueError(f"The source value {source} is not yet implemented. Currently supported sources are 'c3d' "
                             f"and 'dlc' (DeepLabCut)")

        if correct_shift:

            if shift_reference_marker not in self.markers_df.columns.get_level_values("bodyparts").tolist():
                raise ValueError("The shift reference marker " + shift_reference_marker + " is not among the markers."
                                 + "\n Please select one among the following: \n" +
                                 self.markers_df.columns.tolist())

            if not set(to_shift).issubset(self.markers_df.columns.get_level_values("bodyparts").tolist()):
                raise ValueError("Some or all columns to shift are not among the markers. You selected: \n"
                                 + " ,".join(str(x) for x in to_shift)
                                 + "\n Please select them among the following: \n" +
                                 ", ".join(str(x) for x in self.markers_df.columns.tolist()))

            self.markers_df = kinematics_processing.shift_correct(self.markers_df, shift_reference_marker, to_shift)

        if correct_tilt:

            if tilt_reference_marker not in self.markers_df.columns.get_level_values("bodyparts").tolist():
                raise ValueError("The tilt reference marker " + tilt_reference_marker + " is not among the markers."
                                 + "\n Please select one among the following: \n" +
                                 ", ".join(str(x) for x in self.markers_df.columns.tolist()))

            if not set(to_tilt).issubset(self.markers_df.columns.get_level_values("bodyparts").tolist()):
                raise ValueError("Some or all columns to tilt are not among the markers. You selected: \n"
                                 + " ,".join(str(x) for x in to_tilt)
                                 + "\n Please select them among the following: \n" +
                                 ", ".join(str(x) for x in self.markers_df.columns.tolist()))

            self.markers_df = kinematics_processing.tilt_correct(self.markers_df, tilt_reference_marker, to_tilt)

        return

    def filter_marker_df(self, func=None, **kwargs):
        """
        Filters markers dataframe in place using a given function or defaults on a savgol filter with window_length 4 and
        polyorder 2. Pass the arguments to change the defaults.
        :param func: user-given filter function. Must be used as an applied function on a dataset
        :param kwargs: user-given parameters for the savgol filter, or the custom filter function.
        :return:
        """
        if not func:
            func = partial(savgol_filter, window_length=4, polyorder=2)
        original_shape = self.markers_df.shape
        filtered_df = self.markers_df.apply(lambda col: func(col, **kwargs))
        if filtered_df.shape != original_shape:
            raise ValueError(
                f"Shape mismatch, the filter function has to return the same shape as the input. "
                f"Expected {original_shape} returned {filtered_df.shape}")
        self.markers_df = filtered_df

    def convert_DLC_like_to_df(self, multiindex_df=None):
        """
        Converts a DeepLabCut-like dataframe (MultiIndex) to a simple pandas dataframe. Saves the scorer identity and
        the bodyparts in the corresponding class attributes.

        :param smooth: whether to smooth the coordinates or not
        :param filter_window: how big should the filter be
        :param order: order of the polynomial to fit the data
        :return:
        """
        if not multiindex_df:
            multiindex_df = self.markers_df
        bodyparts = multiindex_df.columns.get_level_values("bodyparts").unique().to_list()
        scorers = np.unique(multiindex_df.columns.get_level_values(0))
        scorer = [i for i in scorers if "Unnamed" not in i][0]
        selected_df = multiindex_df.loc[:, multiindex_df.columns.get_level_values("scorer") == scorer]
        selected_df.columns = ["_".join(a[1:]) for a in selected_df.columns.to_flat_index()]
        selected_df.reset_index(inplace=True, drop=True)
        self.markers_df = selected_df
        self.bodyparts = bodyparts
        self.scorer = scorer
        return

    def compute_gait_cycles_bounds(self, left_marker, right_marker, axis="z"):
        """
        Computes the lifting and landing frames of both feet using a left and a right marker, respectively.
        To increase robustness of the cycle estimation it first low-passes the signal.

        :param left_marker: reference marker for the left foot, typically the left mtp
        :param right_marker: reference marker for the right foot, typically the right mtp
        :param recording_fs: sample frequency of the recording, used for low-passing.
        :param axis: axis to use to trace the movement and set the cycle bounds
        :return:
        """

        if left_marker not in self.markers_df.columns.get_level_values("bodyparts").unique():
            raise ValueError("The left reference marker " + left_marker + " is not among the markers."
                             + "\n Please select one among the following: \n" +
                             ", ".join(str(x) for x in self.markers_df.columns.get_level_values("bodyparts").unique()))

        if right_marker not in self.markers_df.columns.get_level_values("bodyparts").unique():
            raise ValueError("The right reference marker " + right_marker + " is not among the markers."
                             + "\n Please select one among the following: \n" +
                             ", ".join(str(x) for x in self.markers_df.columns.get_level_values("bodyparts").unique()))

        self.left_mtp_lift, self.left_mtp_land, self.left_mtp_max = event_detection.get_toe_lift_landing(
            self.markers_df[self.scorer][left_marker][axis], self.fs)
        self.right_mtp_lift, self.right_mtp_land, self.right_mtp_max = event_detection.get_toe_lift_landing(
            self.markers_df[self.scorer][right_marker][axis], self.fs)

        return

    def plot_step_partition(self, step_left, step_right, ax_l, ax_r):
        """
        Plots the step partition of a gait trace (ideally toe or foot trace). Fetches the class attributes so the
        steps computation has to happen before

        :param step_left: name of the left trace body part
        :param step_right: name of the right trace body part
        :param ax_l: ax to plot the left trace on
        :param ax_r: ax to plot the right trace on
        :return: axes
        """
        step_trace_l = self.markers_df[self.scorer][step_left]["z"]
        ax_l.plot(step_trace_l)
        ax_l.vlines(self.left_mtp_lift, min(step_trace_l), max(step_trace_l), colors="green")
        ax_l.vlines(self.left_mtp_land, min(step_trace_l), max(step_trace_l), colors="red")
        ax_l.set_title("Left side")

        step_trace_r = self.markers_df[self.scorer][step_right]["z"]
        ax_r.plot(step_trace_r)
        ax_r.vlines(self.right_mtp_lift, min(step_trace_r), max(step_trace_r), colors="green")
        ax_r.vlines(self.right_mtp_land, min(step_trace_r), max(step_trace_r), colors="red")
        ax_r.set_title("Right side")

        ax_l.set_xlim(0, len(step_trace_l))
        ax_r.set_xlim(0, len(step_trace_r))

        return ax_l, ax_r

    def print_step_partition(self, step_left, step_right, output_folder="./"):
        """
        Creates and saves the step partition plot. See plot_step_partition method for details.

        :param step_left: name of the left trace body part
        :param step_right: name of the right trace body part
        :param output_folder: where to save the figure
        :return:
        """
        fig, axs = plt.subplots(2, 1)
        fig.tight_layout(pad=2.0)
        filename = output_folder + self.path.split("/")[-1] + "_steps_partition.png"
        axs[0], axs[1] = self.plot_step_partition(step_left, step_right, axs[0], axs[1])
        plt.savefig(filename, facecolor="white")
        plt.close()

    def extract_features(self, get_binned=True, custom_feats=None):
        """
        Computes features on the markers dataframe based on the config file. If get_binned is True, it also computes
        the binned features as defined in the config file.

        :return:
        """
        features = self.config["features"]
        skeleton = self.config["skeleton"]
        binning = self.config["binning"]
        new_features, new_binned_features = feature_extraction.extract_features(features=features,
                                                                                bodyparts=self.bodyparts,
                                                                                skeleton=skeleton,
                                                                                markers_df=self.markers_df,
                                                                                get_binned=get_binned,
                                                                                bin_params=binning,
                                                                                custom_feats=custom_feats)

        if self.features_df is not None:
            self.features_df = pd.concat((self.features_df, new_features), axis=1)
        else:
            self.features_df = new_features

        if self.binned_df is not None:
            self.binned_df = pd.concat((self.binned_df, new_binned_features), axis=1)
        else:
            self.binned_df = new_binned_features

    def get_trace_height(self, marker, axis="z", window=50, overlap=25):
        """
        Computes height traces of a given marker

        :param marker: which marker to consider
        :param axis: x, y, or z, default z
        :param window: window to compute the binning on
        :param overlap: how much should 2 subsequent windows overlab by
        :return:
        """
        test = binning.get_step_height_on_bins(self.markers_df,
                                               window=window,
                                               overlap=overlap,
                                               marker=marker,
                                               axis=axis)
        return test

    def get_step_fwd_movement_on_bins(self, marker, axis="y", window=50, overlap=25):
        """
        Computes forward movement traces of a given marker

        :param marker: which marker to consider
        :param axis: x, y, or z, default z
        :param window: window to compute the binning on
        :param overlap: how much should 2 subsequent windows overlab by
        :return:
        """
        test = binning.get_step_fwd_movement_on_bins(self.markers_df,
                                                     window=window,
                                                     overlap=overlap,
                                                     marker=marker,
                                                     axis=axis)
        return test

    def gait_param_to_csv(self, output_folder="./"):
        """
        Writes the gait_param dataframe to a csv file with the name [INPUT_FILENAME]+_gait_param.csv

        :return:
        """
        self.gait_param.to_csv(output_folder + self.path.split("/")[-1].replace(".c3d", "_gait_param.csv"))
        return

    def stepwise_gait_features_to_csv(self, output_folder="./"):
        """
        Writes the gait_param dataframe to a csv file with the name [INPUT_FILENAME]+_gait_param.csv

        :return:
        """
        self.stepwise_gait_features.to_csv(
            output_folder + self.path.split("/")[-1].replace(".c3d", "stepwise_feature.csv"))
        return

    def get_angles_features(self, features_df, **kwargs):
        left_df, right_df = self.split_in_unilateral_df(**kwargs)
        left_features = pd.DataFrame()
        right_features = pd.DataFrame()

        left_features = kinematics_processing.get_angle_features(left_df, self.left_mtp_land,
                                                                 features_df)  # TODO update features df with all angle features

        self.stepwise_gait_features = left_features.join(right_features)



import numpy as np
import pandas as pd


def parse_df_features_for_binning(df_markers, df_features):
    df_markers_no_scorer = df_markers.copy().droplevel("scorer", axis=1)
    cols = ['_'.join(col) for col in df_markers_no_scorer.columns.values]
    df_grouped = df_features.copy()
    df_grouped = df_grouped.droplevel("scorer", axis=1)
    df_grouped.columns = df_grouped.columns.to_flat_index()
    df_grouped.columns = ['_'.join(col) for col in df_grouped.columns.values]
    #test_names = [col.replace(" ", "_") for col in df_grouped.columns.values]
    df_grouped.drop(cols, inplace=True, axis=1)

    return df_grouped

def get_easy_metrics_on_bins(df_grouped, window, overlap):

    means = df_grouped.rolling(window=window, step=overlap).mean().add_suffix("_mean")
    stds = df_grouped.rolling(window=window, step=overlap).std().add_suffix("_std")
    maxs = df_grouped.rolling(window=window, step=overlap).max().add_suffix("_max")
    mins = df_grouped.rolling(window=window, step=overlap).max().add_suffix("_min")
    metrics = pd.concat((means, stds, maxs, mins), axis=1)
    return metrics


def get_step_height_on_bins(df_markers, marker, window, overlap, axis):
    df_markers_no_scorer = df_markers.copy().droplevel("scorer", axis=1)
    trace_column = df_markers_no_scorer[marker, axis]
    metric = trace_column.rolling(window=window, step=overlap).apply(get_step_height)
    pd_metric = metric.to_frame(name=marker+"_height")
    return pd_metric


def get_step_fwd_movement_on_bins(df_markers, marker, window, overlap, axis):
    df_markers_no_scorer = df_markers.copy().droplevel("scorer", axis=1)
    trace_column = df_markers_no_scorer[marker, axis]
    metric = trace_column.rolling(window=window, step=overlap).apply(get_step_len)
    pd_metric = metric.to_frame(name=marker+"_fwd_move")
    return pd_metric


def get_step_height(trace):
    trace = trace + np.min(trace)
    return np.absolute(np.max(trace) - np.min(trace))


def get_step_len(trace):
    return trace.values[-1] - trace.values[0]



from neurokin.utils.neural import importing
from numpy.typing import ArrayLike


class NeuralData:
    """
    This class represents the neural data recording of a single recording experiment.
    It allows to load neural data from multiple recording systems and to perform basic analysis on the raw data.

    """

    def __init__(self, path):
        self.path = path
        self.raw: ArrayLike
        self.fs: float
        self.sync_data: ArrayLike
        self.fs_sync: float
        self.pulses_count: int
        self.recording: int
        self.freq: ArrayLike[ArrayLike]
        self.pxx: ArrayLike[ArrayLike]
        self.recording_duration: float

    def load_tdt_data(self, sync_present: bool = False, stream_name="Wav1", stim_stream_name="Wav1", t1=0, t2=-1):
        """
        Loads TDT data in the classes attributes.

        :param sync_present: whether there is a sync channel or not (e.g. stimulation trace)
        :param stream_name:  name of the stream where the neural data is stored
        :param stim_stream_name: name of the stream where the sync or stimulation data is stored
        :param t1: first time in second from which to import
        :param t2: last time in second up until which to import
        :return:
        """
        self.fs, self.raw, self.sync_data, self.fs_sync = importing.import_tdt_channel_data(folderpath=self.path,
                                                                                            stream_name=stream_name,
                                                                                            stim_name=stim_stream_name,
                                                                                            sync_present=sync_present,
                                                                                            t1=t1,
                                                                                            t2=t2)

    def load_open_ephys(self, experiment, recording, channels=None, sync_present: bool = False, sync_ch: int = None,
                        source_processor: str = None):
        """
        Loads OpenEphys data in the classes attributes

        :param experiment: name of the experiment folder, as named by OE
        :param recording: name of the recording folder, as named by OE
        :param channels: which channels to load
        :param sync_present: whether there is a sync channel or not (e.g. stimulation trace)
        :param sync_ch: number where the sync or stimulation data is stored
        :param source_processor: name of the source processor as named by OE
        :return:
        """
        self.fs, self.raw, self.sync_data = importing.import_open_ephys_channel_data(folderpath=self.path,
                                                                                     experiment=experiment,
                                                                                     recording=recording,
                                                                                     channels=channels,
                                                                                     sync_present=sync_present,
                                                                                     sync_ch=sync_ch,
                                                                                     source_processor=source_processor)
        self.recording = recording


    def compute_recording_duration(self):
        """
        Computes recording duration based on len of the dataset over the sampling frequency and sets it as a class
        attribute

        :return:
        """
        self.recording_duration = self.raw.shape[1] / self.fs


    def pick_sync_data(self, idx):
        """
        In TDT loading, if the sync_data is accidentally a multichannel, use this to pick the correct one, based on
        a 0-based index.

        :param idx: index of the correct sync data
        :return:
        """
        self.sync_data = self.sync_data[idx]



import numpy as np
from numpy.typing import ArrayLike
from typing import List
from matplotlib import axes as ax


def plot_spectrogram(ax: ax, fs: float, raw: ArrayLike, ylim: List[int], title: str = "", tick_spacing: int = 10,
                     nfft: int = None, noverlap: int = None, vmin=None, vmax=None, **kwargs) -> ax:
    """
    Wrapper for matplotlib spectrogram to add some style elements.

    :param ax: axes where to plot
    :param fs: sampling frequency
    :param raw: raw data
    :param title: title of the figure
    :param ylim: limit of frequencies
    :param tick_spacing: spacing of the ticks on the axis
    :param nfft: nfft to comput the fft
    :param noverlap: number of overlap points
    :return:
    """
    ax.set_title(title)
    ax.specgram(raw, NFFT=nfft, Fs=fs, noverlap=noverlap, vmin=vmin, vmax=vmax, **kwargs)
    ax.set_ylim(ylim)
    ax.set_yticks(np.arange(ylim[0], ylim[1], tick_spacing))
    return ax


def plot_welch(ax: ax, freqs: ArrayLike, pxx_den: ArrayLike, title: str = "PSD", xlim=None,
               ylim=None,
               tick_spacing: int = 10) -> ax:
    """
    Returns a plot of psd with some style elements

    :param ax: axes where to plot
    :param freqs: array like of frequencies corresponding to the x axis
    :param pxx_den: array like of powers corresponding to the y axis
    :param title: str
    :param xlim: limit of frequencies for plotting
    :param ylim: limit of power for plotting
    :param tick_spacing:  spacing of the ticks on the axis
    :return:
    """
    ax.set_title(title)

    if not xlim is None:
        ax.set_xlim(xlim)
        ticks_min = xlim[0]
        ticks_max = xlim[1]
    else:
        ticks_min = min(freqs)
        ticks_max = max(freqs)

    if not ylim is None:
        ax.set_ylim(ylim)

    ax.set_xticks(np.arange(ticks_min, ticks_max, tick_spacing))

    ax.set_xlabel('frequency [Hz]')
    ax.set_ylabel('PSD [V**2/Hz]')

    ax.semilogy(freqs, pxx_den)
    return ax



import tdt
import json
import numpy as np
from typing import Union
from numpy.typing import ArrayLike
from neurokin.constants.open_ephys_structure import STRUCTURE, CONTINUOUS, SOURCE_PROCESSOR_NAME, SOURCE_PROCESSOR_ID, \
    TRAILING_NUMBER, SAMPLE_RATE, CHANNEL_NUMBER


def time_to_sample(timestamp: float, fs: float, is_t1: bool = False, is_t2: bool = False) -> int:
    """
    Function adapted from time2sample in TDTbin2py.py
    Returns the sample index given a time in seconds and the sampling frequency.
    It has to be specified if the timestamp refers to t1 or t2.
    :param timestamp: time in seconds
    :param fs: sampling frequency
    :param is_t1: specify if the timestamp is t1
    :param is_t2: specify if the timestamp is t2
    :return:
    """
    if timestamp < 0:
        raise ValueError("Timestamp cannot be negative")
    if not fs > 0:
        raise ValueError("Sampling frequency must be positive")
    sample = timestamp * fs
    if is_t2:
        exact = np.round(sample * 1e9) / 1e9
        sample = np.floor(sample)
        if exact == sample:
            sample -= 1
    else:
        if is_t1:
            sample = np.ceil(sample)
        else:
            sample = np.round(sample)
    sample = int(sample)
    if sample < 0:
        sample = 0
    return sample


def import_tdt_channel_data(folderpath, ch=[], t1=0, t2=-1, stream_name="Wav1", stim_name="Wav1",
                            sync_present=False) -> (
        float, ArrayLike):
    """
    Wrapper for the import function of tdt, to be more user friendly.


    :param folderpath: folderpath of the subject experiment
    :param ch:
    :param stream_name:
    :param stim_name:
    :param sync_present:
    :param t1: initial time to index in seconds
    :param t2: last time to index in seconds
    :return: frequency sample and raw sample
    """
    if isinstance(ch, (list, np.ndarray)):
        if len(ch) > 0:
            ch = [i + 1 for i in ch]
        if len(ch) == 0:
            ch = 0
    elif type(ch) == int:
        ch += 1

    data = tdt.read_block(folderpath, evtype=['streams'], channel=ch)
    stim_data = None
    fs_sync = None
    try:
        streams = data.streams
        stored = getattr(streams, stream_name)
        raw = stored.data
        fs = stored.fs

    except AttributeError:
        print(f"No stream named  {stream_name}, please specify the correct stream_name \n "

              f"Please chose from: {data.streams.__dict__.keys()}")

        return

    s1 = 0
    s2 = -1
    if t1 != 0:
        s1 = time_to_sample(timestamp=t1, fs=fs, is_t1=True)
    if t2 != -1:
        s2 = time_to_sample(timestamp=t2, fs=fs, is_t2=True)
    raw = raw[..., s1:s2]

    if sync_present:
        try:
            streams = data.streams
            stored = getattr(streams, stim_name)
            stim_data = stored.data
            fs_sync = stored.fs

        except AttributeError:
            print(f"No stimulation data named {stim_name}, please specify the correct stream_name \n "
                  f"Please chose from: {data.streams.__dict__.keys()}")
            return
        stim_data = stim_data[..., s1:s2]

    if fs_sync is not None:
        if fs_sync != fs:
            print("Warning: sync frequency is different from the data frequency")

    return fs, raw, stim_data, fs_sync


def import_open_ephys_channel_data(folderpath: str, experiment: str, recording: str,
                                   channels: Union[list, int] = None,
                                   sync_present: bool = False,
                                   sync_ch: int = None,
                                   source_processor: str = None,
                                   start_sample_index: int = 0,
                                   end_sample_index: int = -1,
                                   convert_to_volts: bool = True) -> (
        float, np.ndarray):
    """
    Imports open ephys data from binary files.

    :param folderpath: Folderpath where the experiment is, including the Node
    :param experiment: experiment folder
    :param recording: recording folder
    :param channels: indicate which channels to return
    :return: Sampling frequency is returned as a float and raw data are returned in arbitrary units
    """

    structure_path = folderpath + "/" + experiment + "/" + recording + "/" + STRUCTURE + ".oebin"
    sync_data = None

    with open(structure_path) as f:
        structure = json.load(f)

    if source_processor is None:
        source_processor = str(structure[CONTINUOUS][0][SOURCE_PROCESSOR_NAME].replace(" ", "_")) + "-" + \
                           str(structure[CONTINUOUS][0][SOURCE_PROCESSOR_ID]) + TRAILING_NUMBER

    binary_data_path = folderpath + "/" + experiment + "/" + recording + "/" + \
                       CONTINUOUS + "/" + source_processor + "/" + CONTINUOUS + ".dat"

    fs = structure[CONTINUOUS][0][SAMPLE_RATE]
    n_ch = structure[CONTINUOUS][0][CHANNEL_NUMBER]
    bit_volts = [ch['bit_volts'] for ch in structure["continuous"][0]['channels']]

    # neural_data_flat = np.fromfile(binary_data_path, dtype='<i2')
    # samples = samples[start_sample_index:end_sample_index, channels].astype('float64')
    # samples = data.reshape((len(data) // n_ch, n_ch))
    # neural_data_au = np.reshape(a=neural_data_flat, newshape=(n_ch, n_samples), order='F')
    # if (isinstance(channels, (list, np.ndarray)) and len(channels) > 0) or isinstance(channels, int):
    #    mask = np.zeros(n_ch, dtype=bool)
    #    mask[channels] = True
    #    neural_data_au = samples[mask, ...]
    #    if neural_data_au.shape[0] == 1:
    #        neural_data_au = neural_data_au[0]

    data = np.memmap(binary_data_path, mode='r', dtype='int16')

    n_samples = int(len(data) / n_ch)
    if not channels:
        channels = range(n_ch)
    if type(channels) == int:
        channels = [channels]
    samples = np.reshape(a=data, newshape=(n_ch, n_samples), order='F')
    samples = samples[channels, start_sample_index:end_sample_index].astype('float64')

    if convert_to_volts:
        for i, ch in enumerate(channels):
            samples[i] = samples[i] * bit_volts[ch]

    if sync_present:
        sync_data = samples[sync_ch]

    return fs, samples, sync_data


# TESTME
def import_binary_to_float32(filename, channel_number, sample_number):
    """
    Imports binary data stored in C major to a float32 array

    :param filename: file to import from
    :param sample_number: number of sample in each channel
    :param channel_number: number of channels
    :return: the array with shape channel_number*sample_number
    """
    dt = np.dtype('f4')
    test = np.fromfile(filename, dtype=dt)
    float_array = np.reshape(a=test, newshape=(channel_number, sample_number), order='C')
    return float_array



from typing import Tuple

import numpy as np
from numpy import ndarray
from scipy import signal

from neurokin.constants.gait_cycle_detection import RELATIVE_HEIGHT, STEP_FILTER_FREQ, PROMINENCE


# TESTME with pickled steps data
def get_toe_lift_landing(y, recording_fs):
    """
    Returns the left and right bounds of the gait cycle, corresponding to the toe lift off and the heel strike.
    As a first step it smooths the signal to increase robust peak detection.

    :param y: trace of the toe in the z coordinate
    :param recording_fs: sampling rate of the kinematics recording
    :return: left, right bounds and max value of each peak.
    """
    y = lowpass_array(y, STEP_FILTER_FREQ, recording_fs)

    max_x, _ = signal.find_peaks(y, prominence=PROMINENCE)
    avg_distance = abs(int(median_distance(max_x) / 2))

    lb = []
    rb = []

    for p in max_x:
        left = p - avg_distance if p - avg_distance > 0 else 0
        right = p + avg_distance if p + avg_distance < len(y) else len(y)
        bounds = get_peak_boundaries_scipy(y=y[left:right], px=p, left_crop=left)
        lb.append(bounds[0])
        rb.append(bounds[1])

    lb = np.asarray(lb)
    rb = np.asarray(rb)
    return lb, rb, max_x


# TESTME with pickled steps data
def get_peak_boundaries_scipy(y: ndarray, px: float, left_crop: int) -> Tuple[int, int]:
    """
    Computes the boundaries of a step by getting the peaks boundaries

    :param y: y values
    :param px: peaks indexes
    :param left_crop: how many samples to use to crop on the left side
    :return: returns boundaries of steps
    """
    peaks = np.asarray([px - left_crop])
    peak_pro = signal.peak_prominences(y, peaks)
    peaks_width = signal.peak_widths(y, peaks, rel_height=RELATIVE_HEIGHT, prominence_data=peak_pro)
    intersections = peaks_width[-2:]
    try:
        left = int(intersections[0] + left_crop)
    except:
        left = int(left_crop)

    try:
        right = int(intersections[-1] + left_crop)
    except:
        right = int(len(y) + left_crop)

    return left, right


# TESTME
def lowpass_array(array, critical_freq, fs):
    """
    Low passes the array for a given frequency using a 2nd order butterworth filter and filtfilt to avoid phase shift.

    :param array: input array
    :param critical_freq: critical filter frequency
    :param fs: sampling frequency
    :return: filtered array
    """
    b, a = signal.butter(2, critical_freq, "low", output="ba", fs=fs)
    filtered = signal.filtfilt(b, a, array)
    return filtered


# TESTME
def median_distance(a: ndarray) -> ndarray:
    """
    Gets median distance between peaks

    :param a:
    :return: median
    """
    distances = []
    for i in range(len(a) - 1):
        distances.append(a[i] - a[i + 1])
    return np.median(distances)



import os
import re
from typing import List, Dict
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pickle as pkl
from neurokin.utils.experiments.neural_correlates import (get_events_dict, get_neural_correlates_dict, compute_psd_for_row)
from neurokin.utils.experiments.neural_states_helper import (
                                                             get_per_animal_psds_df, save_data,
                                                             compute_events_percentage,
                                                             condense_distribution_event_types,
                                                             get_group_split, get_state_graph_stats,
                                                             get_runs_list)
from neurokin.utils.helper.load_config import read_config
from neurokin.utils.experiments.neural_correlates_plot import plot_psd_single_state


class NeuralCorrelatesStates():
    """
    This class allows to deal with unimodal and multimodal retrieval of sata based on states timestamps.
    """
    def __init__(self,
                 timeslice: float,
                 experiment_structure_filepath: str,
                 skip_subjects: List[str] = [],
                 skip_conditions: List[str] = [],
                 skiprows: int = 2,
                 framerate: float = 200):

        self.timeslice = timeslice
        self.experiment_structure = read_config(experiment_structure_filepath, converts_keys_to_string=True)
        self.skip_subjects = skip_subjects
        self.skip_conditions = skip_conditions
        self.skiprows = skiprows
        self.framerate = framerate
        self.stream_names: List[str] = []
        self.ch_of_interest: Dict[str, int] = {}
        self.freqs: np.array = None
        self.fs: float = None
        self.events_dataset: pd.DataFrame = None
        self.raw_neural_correlates_dataset: pd.DataFrame = None
        self.psds_correlates_dataset: pd.DataFrame = None

    def save_dataset(self, dataset, filename):
        """
        Saves dataset dictionary to pickle file

        :param dataset: str name of which dataset to save
        :param filename: filename without extension
        :return:
        """
        accepted_datasets = ["events_dataset", "raw_neural_correlates_dataset", "psd_neural_correlates_dataset"]
        if dataset not in accepted_datasets:
            raise ValueError(f"Dataset not found. Please select one of the following : {accepted_datasets}")

        if dataset == "events_dataset":
            save_data(self.events_dataset, filename)
        if dataset == "raw_neural_correlates_dataset":
            save_data(self.fs, "fs")
            save_data(self.raw_neural_correlates_dataset, filename)
        if dataset == "psd_neural_correlates_dataset":
            save_data(self.psds_correlates_dataset, filename)
            save_data(self.freqs, "freqs")

    def create_events_dataset(self, experiment_path, verbose=False, file_starts_with=None):
        """
        Takes the experiment structure and based on the runs listed there it looks for .csv files to import.
        Then based on the labels, creates a dataframe containing condition, date, animal, run and event.
        Each event contains the list of timestamps [start, end] for that specific run

        :param experiment_path: folder where to find the dataset
        :param verbose: if True it will print the currently processed run
        :param file_starts_with: set if the files to fetch begin with a specific pattern
        :return: dataframe containing data of each trial  and timestamps of events
        :param experiment_path:
        """
        trial_list = get_runs_list(self.experiment_structure, self.skip_subjects, self.skip_conditions)
        trial_data = []
        for trial in trial_list:
            date, subject, condition, run = trial
            run_path = "/".join([experiment_path, date, subject, run]) + "/"

            if verbose:
                print(f"Currently processing: {date} - {subject} - {condition} - {run}")
            try:
                if file_starts_with:
                    event_path = [run_path + fname for fname in os.listdir(run_path)
                                  if fname.startswith(file_starts_with) and fname.endswith(".csv")][0]
                else:

                    event_path = [run_path + fname for fname in os.listdir(run_path) if
                                  re.match(r"(?i)[a-z_-]+[0-9]{1,3}.csv", fname)][0]
                events_dict = get_events_dict(event_path=event_path,
                                              skiprows=self.skiprows,
                                              framerate=self.framerate)
                events = [events_dict["gait"],
                          events_dict["nlm_rest"],
                          events_dict["nlm_active"],
                          events_dict["fog_rest"],
                          events_dict["fog_active"]]

                trial_data.append(trial + events)

            except FileNotFoundError as error:
                print(f"{error} No .csv events File Found For {date}, {subject}, {run}")

        df = pd.DataFrame(trial_data, columns=["date",
                                               "subject",
                                               "condition",
                                               "run",
                                               "event_gait",
                                               "event_nlm_rest",
                                               "event_nlm_active",
                                               "event_fog_rest",
                                               "event_fog_active"])
        self.events_dataset = df
        return

    def create_raw_neural_dataset(self, experiment_path, stream_names: List[str], ch_of_interest: Dict[str, int],
                                  verbose=False):
        """
        Creates a pandas dataframe containing neural chunks corresponding to locomotor events in each run.

        :param experiment_path: path to the dataset
        :param stream_names: names of streams where the data is saved (TDT specific)
        :param ch_of_interest: channel of interest. Neural chunks are for only one channel
        :param verbose: whether to print info on the run currently processed
        :return:
        """

        if self.events_dataset is None:
            print("Please create or load an events dictionary first, "
                  "using either the method create_events_dataset or load_dataset"
                  "If the dataset is loaded it should be assigned to the attribute events_dataset_dict")

            return
        self.stream_names = stream_names
        self.ch_of_interest = ch_of_interest
        trial_list = get_runs_list(self.experiment_structure, self.skip_subjects, self.skip_conditions)
        trial_neural_data = []
        for trial in trial_list:
            date, subject, condition, run = trial
            run_path = "/".join([experiment_path, date, subject, run]) + "/"
            channel_of_interest = self.ch_of_interest[subject]
            neural_event_df = self.events_dataset[(self.events_dataset["date"] == date) &
                                                  (self.events_dataset["subject"] == subject) &
                                                  (self.events_dataset["condition"] == condition) &
                                                  (self.events_dataset["run"] == run)]
            if verbose:
                print(f"Currently processing: {date} - {subject} - {condition} - {run}")

            raw_neural_correlate, fs = get_neural_correlates_dict(neural_path=run_path,
                                                                  channel_of_interest=channel_of_interest,
                                                                  stream_names=stream_names,
                                                                  events_df=neural_event_df,
                                                                  time_cutoff=self.timeslice)
            if fs is not None:
                self.fs = fs

            events = [raw_neural_correlate["event_gait"],
                      raw_neural_correlate["event_nlm_rest"],
                      raw_neural_correlate["event_nlm_active"],
                      raw_neural_correlate["event_fog_rest"],
                      raw_neural_correlate["event_fog_active"]]

            trial_neural_data.append(trial + events + [fs])
        columns = self.events_dataset.columns.tolist()
        columns.append("fs")
        df = pd.DataFrame(trial_neural_data, columns=columns)
        self.raw_neural_correlates_dataset = df

    def create_psd_dataset(self, nfft, nov, zscore=False):
        """
        Computes the Power Spectra Density of the neural correlates.

        :param nfft: NFFT parameter to use for the Fourier Transform
        :param nov: Overlap parameter to use for the Fourier Transform
        :param verbose: if True it will print the currently processed run
        :return: dictionary structured as condition, date, animal, run, event containing the PSD neural correlates.
        """
        if self.raw_neural_correlates_dataset is None:
            print("Please create or load a raw neural dictionary first, "
                  "using either the method create_events_dataset or load_dataset."
                  "If the dataset is loaded it should be assigned to the attribute raw_neural_correlates_dict")
            return
        events_columns = [c for c in self.raw_neural_correlates_dataset.columns if c.startswith("event")]
        meta_columns = [c for c in self.raw_neural_correlates_dataset.columns if not c.startswith("event")]
        psds_correlates_dataset = self.raw_neural_correlates_dataset.apply(lambda row: compute_psd_for_row(row,
                                                                                                           events_columns,
                                                                                                           nfft,
                                                                                                           nov,
                                                                                                           zscore), axis=1)
        #psds_correlates_dataset = self.raw_neural_correlates_dataset[events_columns].applymap(
        #    get_psd_single_event_type,
        #    fs=self.fs,
        #    nfft=nfft,
        #    noverlap=nov,
        #    zscore=zscore)
        self.psds_correlates_dataset = pd.concat((self.raw_neural_correlates_dataset[meta_columns],
                                                  psds_correlates_dataset), axis=1)
        #self.freqs = np.fft.rfftfreq(n=nfft, d=1 / self.fs)


    def plot_prep_psds_dataset(self, test_sbj_list, condense=True):
        """
        Creates a reduced dataframe of PSD ready to be plotted

        :param test_sbj_list: list of subjects that belong in the experimental condition
        :param condense: whether to retain the 5 states format or condense to the 3 format.
        :return:
        """
        per_animal_avg = get_per_animal_psds_df(self.psds_correlates_dataset, condense=condense)
        group_split = get_group_split(test_sbj_list=test_sbj_list, df=per_animal_avg)
        return group_split


    def plot_prep_states_distribution(self, test_sbj_list, condense=True, stat="std"):
        """
        Fixed shortcut to generate a stats dictionary of the state distribution, ready to be plotted

        :param test_sbj_list: list of subject IDs that belong to the test group
        :param condense:  bool if to condense from 5 categories to 3
        :return: stats dictionary dataset structured as group, condition, state, stats {"mean":, "upper_bound", "lower_bound}
        """
        events_percentage = compute_events_percentage(self.events_dataset)
        if condense:
            events_percentage = condense_distribution_event_types(events_percentage)
        group_split = get_group_split(test_sbj_list=test_sbj_list, df=events_percentage)
        stats = get_state_graph_stats(group_cond_df=group_split, stat=stat)
        return stats


if __name__ == "__main__":
    NFFT = 2 ** 12
    NOV = int(NFFT / 4)
    TIME_CUTOFF = 1.5
    experiment_structure_path = "../../../analysis/neural_correlates_states_clean/experiment_structure.yaml"
    pda = ["NWE00053", "NWE00054", "NWE00130", "NWE00160", "NWE00161", "NWE00162", "NWE00163", "NWE00164"]
    CHANNEL_DICT = {"NWE00052": 6,
                    "NWE00053": 1,
                    "NWE00054": 1,
                    "NWE00089": 1,
                    "NWE00090": 1,
                    "NWE00092": 1,
                    "NWE00093": 1,
                    "NWE00130": 3,
                    "NWE00131": 2,
                    "NWE00158": 3,
                    "NWE00159": 3,
                    "NWE00166": 3,
                    "NWE00160": 3,
                    "NWE00161": 3,
                    "NWE00162": 3,
                    "NWE00163": 3,
                    "NWE00164": 3}
    skip_animals = [i for i in CHANNEL_DICT.keys() if not i=="NWE00130"]

    ncs = NeuralCorrelatesStates(timeslice=TIME_CUTOFF,
                                 experiment_structure_filepath=experiment_structure_path,
                                 skip_subjects=skip_animals)
    ncs.events_dataset_dict = ncs.events_dataset = pd.read_pickle("../../../analysis/neural_correlates_states_clean/raw_states.pkl")

    #ncs.fs = 24414.1
    #ncs.raw_neural_correlates_dataset = pd.read_pickle(
    #    "../../../analysis/neural_correlates_states_clean/raw_neural.pkl")
    #ncs.create_events_dataset("../../../analysis/neural_correlates_states_clean/data/neural_correlates_cervical_dataset/")
    ncs.create_raw_neural_dataset("../../../analysis/neural_correlates_states_clean/data/neural_correlates_cervical_dataset/",
                                  stream_names=["LFP1", "NPr1", "EOG1"],
                                  ch_of_interest=CHANNEL_DICT)
    with open('../../../analysis/neural_correlates_states_clean/data/neural_correlates_cervical_dataset/raw_neural_fs.pkl', 'wb') as handle:
        pkl.dump(ncs.raw_neural_correlates_dataset, handle, protocol=pkl.HIGHEST_PROTOCOL)
    ncs.create_psd_dataset(NFFT, NOV, zscore=False)
    df = ncs.plot_prep_psds_dataset(test_sbj_list=pda, condense=True)

    fig, ax = plt.subplots()
    plot_psd_single_state(ax, df, group=True,
                                condition="baseline",
                                state="event_gait",
                                freqs=ncs.freqs,
                                color="crimson",
                                idx_min=2,
                                idx_max=75)

    print("")
